#!/bin/bash

# build-chroot.sh v.0.1
# $Id: build-chroot.sh 691 2017-06-04 23:12:04Z fwbergm $
# FWB 2013-07-21
# http://www.tuxad.com

usage() {
  cat <<EOF

Build a chroot directory with base packages

usage:
build-chroot.sh [options] RPMDIR CHROOTDIR
examples:
build-chroot.sh --noscripts /rhel64/Packages /home/chroot/rhel64-chroot

options:
--nodeps      option passed through to rpm command
--noscripts   option passed through to rpm command
--nomd5       option passed through to rpm command
--setbase1    install only set base1
--setbase2    install base1 and base2
--setboot     install base1, base2 and boot
--setbase3    install all base sets and boot
--setrpm      install all base sets, boot and rpm
--norpm       install w/o rpm db (rpm2cpio needed)
--bootstrap   first create rpm bootstrap and use this to install (for cross distro build)

This script installs a minimal chroot environment by a given directory with rpm files.
Internally this script has a couple of dependent sets of rpm packages (ordered list):

  base1     filesystem, libc, bash
  base2     coreutils, ssl, pam, python
  boot      udev, ip, syslog, initscripts, dm
  base3     mkinitrd, lvm, raid, dm-multipath
  rpm       rpm, yum
  optional  ...

On trouble like i.e. rpm db trouble when building cross different distro
versiones (i.e. RHEL 6 chroot on RHEL 5 system) you should try bootstrapping.
Building chroot with incompatible archs can only be done without rpm but
with rpm2cpio.

EOF
}

[ $# -lt 2 ] && { usage; exit 1; }

unset RPMDIR
unset CHROOTDIR
unset NODEPS
unset NOSCRIPTS
unset NOMD5
unset NORPM
unset BOOTSTRAP
unset MOUNT
# default ordered pkg set
PKGSET="12b3ro"
while true
do
  case $1 in
    -h)
      usage
      exit 0
      ;;
    --help)
      usage
      exit 0
      ;;
    --nodeps)
      NODEPS="--nodeps"
      shift
      ;;
    --noscripts)
      NOSCRIPTS="--noscripts"
      shift
      ;;
    --nomd5)
      NODEPS="--nomd5"
      shift
      ;;
    --setbase1)
      PKGSET="1"
      shift
      ;;
    --setbase2)
      PKGSET="12"
      shift
      ;;
    --setboot)
      PKGSET="12b"
      shift
      ;;
    --setbase3)
      PKGSET="12b3"
      shift
      ;;
    --setrpm)
      PKGSET="12b3r"
      shift
      ;;
    --norpm)
      NORPM="norpm"
      shift
      ;;
    --bootstrap)
      BOOTSTRAP="bootstrap"
      shift
      ;;
    --mount)
      MOUNT="mount"
      shift
      ;;
    -*)
      echo ERROR: unknown option
      exit 2
      ;;
    *)
      [[ -z $1 ]] && break
      if [[ -z $RPMDIR ]]
      then
        RPMDIR=$1
      else
        if [[ -z $CHROOTDIR ]]
        then
          CHROOTDIR=$1
        else
          echo ERROR: too many arguments
          exit 3
        fi
      fi
      shift
      ;;
  esac
done
[[ -z $RPMDIR ]] && { usage; exit 4; }
[[ -z $CHROOTDIR ]] && { usage; exit 5; }
RPMOPTS="--nomd5 --nosignature $NODEPS $NOSCRIPTS $NOMD5"


mkdir -p "$CHROOTDIR"
pushd "$CHROOTDIR" >/dev/null || { echo ERROR: cannot cd to $CHROOTDIR, exiting; exit 2; }
CHROOTDIR=$(pwd)
/bin/rm -rf *
popd >/dev/null

cd $RPMDIR || { echo ERROR: cannot cd to $RPMDIR, exiting; exit 2; }



# ordered archs, first match wins:
ARCHS_ACCEPTED="
  noarch
  x86_64
  i686
  i386
"

PACKAGES_BASE1_RHEL5="
  filesystem
  basesystem
  redhat-release
  setup
  tzdata
  libgcc
  glibc-common
  glibc
  mktemp
  termcap
  libtermcap
  bash
  libattr
  zlib
  ncurses
  info
  pcre
  libstdc++
  grep
  readline
  db4
  gawk
  libcap
"
PACKAGES_BASE1_RHEL6_ADD="
  redhat-release-server
  ncurses-base
  ncurses-libs
  nss-softokn-freebl
"

PACKAGES_BASE2_RHEL5="
  sed
  libacl
  findutils
  libsepol
  libselinux
  mcstrans
  audit-libs
  shadow-utils
  bzip2-libs
  expat
  keyutils-libs
  krb5-libs
  openssl
  e2fsprogs-libs
  e2fsprogs
  gdbm
  sqlite
  python-libs
  python
  cracklib
  cracklib-dicts
  pam
  popt
  libutempter
  util-linux
  coreutils
"
PACKAGES_BASE2_RHEL6_ADD="
  ca-certificates
  libblkid
  libuuid
  util-linux-ng
  gmp
  coreutils-libs
  libcom_err
  libss
  nss-softokn
  libffi
"

PACKAGES_BOOT_RHEL5="
  MAKEDEV
  udev
  device-mapper
  chkconfig
  iproute
  iputils
  psmisc
  net-tools
  procps
  rsyslog
  logrotate
  mingetty
  glib2
  ethtool
  SysVinit
  module-init-tools
  initscripts
"
PACKAGES_BOOT_RHEL6_ADD="
  hwdata
  libusb
  device-mapper-libs
  sysvinit-tools
  libnih
  dbus-libs
  upstart
  libudev
  diffutils
  checkpolicy
  libselinux-utils
  dbus-glib
  ustr
  libsemanage
  policycoreutils
  iptables
  libidn
  gamin
"

PACKAGES_BASE3_RHEL5="
  mkinitrd
  binutils
  cpio
  libsysfs
  device-mapper-multipath
  sgpio
  dmraid-events
  dmraid
  less
  gzip
  nspr
  nss
  hmaccalc
  iscsi-initiator-utils
  kpartx
  device-mapper-event
  lvm2
  nash
  tar
"
PACKAGES_BASE3_RHEL6_ADD="
  libaio
  device-mapper-multipath-libs
  groff
  nss-sysinit
  nss-util
  device-mapper-event-libs
  device-mapper-persistent-data
  lvm2-libs
"

PACKAGES_RPM_RHEL5="
  libxml2
  m2crypto
  python-elementtree
  python-iniparse
  python-sqlite
  python-urlgrabber
  crontabs
  elfutils-libelf
  rpm-libs
  rpm
  rpm-python
  yum
  yum-metadata-parser
"
PACKAGES_RPM_RHEL6_ADD="
  xz-libs
  file-libs
  libcurl
  python-pycurl
  cyrus-sasl-lib
  cyrus-sasl
  mysql-libs
  nss-tools
  openldap
  postfix
  cronie-anacron
  cronie
  curl
  db4-utils
  libgcrypt
  pinentry
  gnupg2
  libgpg-error
  pth
  gpgme
  pygpgme
  lua
  libssh2
"

PACKAGES_OPTIONAL_RHEL5="
  rootfiles
  vim-minimal
"
#  vim-common

RPM=rpm
RELEASE_FILE="*redhat-release*-[56].*.rpm"
RHEL_RELEASE=$($RPM -qp --qf "%{V}\n" $RELEASE_FILE 2>/dev/null|cut -b-1)
if [[ $RHEL_RELEASE = "6" ]]
then
  echo detected RHEL 6 packages
else
  echo assuming RHEL 5 packages
fi
INSTALLED_RELEASE_PKG=$(rpm -qa --nomd5 --nosignature|grep release|sort|egrep 'fedora|redhat'|head -n 1)
SYS_RELEASE=$($RPM -q --qf "%{V}\n" $INSTALLED_RELEASE_PKG --nomd5 --nosignature 2>/dev/null|cut -b-1)
echo system release is $SYS_RELEASE
[[ $RHEL_RELEASE -ne $SYS_RELEASE ]] && \
  { [[ -n $BOOTSTRAP ]] || \
  echo WARNING: system release differs from $RHEL_RELEASE, maybe you should bootstrap; }

PACKAGES_SELECTED=""
if [[ $PKGSET =~ 1 ]]
then
  PACKAGES_SELECTED="$PACKAGES_BASE1_RHEL5"
  [[ $RHEL_RELEASE = "6" ]] && PACKAGES_SELECTED="$PACKAGES_SELECTED $PACKAGES_BASE1_RHEL6_ADD"
fi
if [[ $PKGSET =~ 2 ]]
then
  PACKAGES_SELECTED="$PACKAGES_SELECTED $PACKAGES_BASE2_RHEL5"
  [[ $RHEL_RELEASE = "6" ]] && PACKAGES_SELECTED="$PACKAGES_SELECTED $PACKAGES_BASE2_RHEL6_ADD"
fi
if [[ $PKGSET =~ b ]]
then
  PACKAGES_SELECTED="$PACKAGES_SELECTED $PACKAGES_BOOT_RHEL5"
  [[ $RHEL_RELEASE = "6" ]] && PACKAGES_SELECTED="$PACKAGES_SELECTED $PACKAGES_BOOT_RHEL6_ADD"
fi
if [[ $PKGSET =~ 3 ]]
then
  PACKAGES_SELECTED="$PACKAGES_SELECTED $PACKAGES_BASE3_RHEL5"
  [[ $RHEL_RELEASE = "6" ]] && PACKAGES_SELECTED="$PACKAGES_SELECTED $PACKAGES_BASE3_RHEL6_ADD"
fi
if [[ $PKGSET =~ r ]]
then
  PACKAGES_SELECTED="$PACKAGES_SELECTED $PACKAGES_RPM_RHEL5"
  [[ $RHEL_RELEASE = "6" ]] && PACKAGES_SELECTED="$PACKAGES_SELECTED $PACKAGES_RPM_RHEL6_ADD"
fi
if [[ $PKGSET =~ o ]]
then
  PACKAGES_SELECTED="$PACKAGES_SELECTED $PACKAGES_OPTIONAL_RHEL5"
fi
PACKAGES_ALL="$PACKAGES_BASE1_RHEL5 $PACKAGES_BASE1_RHEL6_ADD $PACKAGES_BASE2_RHEL5 $PACKAGES_BASE2_RHEL6_ADD $PACKAGES_BOOT_RHEL5 $PACKAGES_BOOT_RHEL6_ADD $PACKAGES_BASE3_RHEL5 $PACKAGES_BASE3_RHEL6_ADD $PACKAGES_RPM_RHEL5 $PACKAGES_RPM_RHEL6_ADD $PACKAGES_OPTIONAL_RHEL5"

PATTERN=$(echo $PACKAGES_SELECTED|sed 's,^,^,;s, ,-|^,g;s,\+,\\+,g;s,$,-,')
OLD_IFS="$IFS"
IFS=""
PACKAGES_AVAILABLE="$(ls|grep "\.rpm$"|egrep "$PATTERN"|xargs $RPM -qp --qf '%{NAME} %{NAME}-%{VERSION}-%{RELEASE}.%{ARCH}\n' 2>/dev/null)"

PACKAGES_TO_INSTALL=""
IFS=$'\n\t '
for PKG in $(echo $PACKAGES_SELECTED)
do
  FOUND_PKG=$(IFS=$'\n\t ';for ARCH in $(echo $ARCHS_ACCEPTED)
    do
      IFS=""
      echo $PACKAGES_AVAILABLE|grep ^"$PKG "$PKG"-.*\."$ARCH"$"
    done|head -n 1|cut -d" " -f2
  )
  if [ "$FOUND_PKG" ]
  then
    PACKAGES_TO_INSTALL="$PACKAGES_TO_INSTALL $FOUND_PKG".rpm
  else
    echo WARNING: no package \"$PKG\" found >&2
  fi
done
IFS=$'\n\t '


if [[ -n $BOOTSTRAP ]]
then
  BOOTSTRAP_CHROOTDIR="$CHROOTDIR".bootstrap
  /bin/rm -rf "$BOOTSTRAP_CHROOTDIR"
  mkdir "$BOOTSTRAP_CHROOTDIR"
  echo search for bootstrap pkgs
  PACKAGES_BOOTSTRAP=""
  IFS=$'\n\t '
  for PKG in $(echo $PACKAGES_ALL)
  do
    FOUND_PKG=$(IFS=$'\n\t ';for ARCH in $(echo $ARCHS_ACCEPTED)
      do
        IFS=""
        echo $PACKAGES_AVAILABLE|grep ^"$PKG "$PKG"-.*\."$ARCH"$"
      done|head -n 1|cut -d" " -f2
    )
    [[ -n $FOUND_PKG ]] && PACKAGES_BOOTSTRAP="$PACKAGES_BOOTSTRAP $FOUND_PKG".rpm
  done
  IFS=$'\n\t '
  echo bootstrapping
  for PKG in $PACKAGES_BOOTSTRAP
  do
    if rpm2cpio $PKG |file -|grep -q xz.compressed
    then
      rpm2cpio $PKG | xz -d
    else
      rpm2cpio $PKG
    fi \
      | ( cd "$BOOTSTRAP_CHROOTDIR" && cpio -i --make-directories --preserve-modification-time --no-absolute-filenames --unconditional 2>/dev/null )
  done
  echo trying to detect loader
  export LDSO="$BOOTSTRAP_CHROOTDIR"/$(readelf -l "$BOOTSTRAP_CHROOTDIR"/bin/rpm|grep "program interpreter"|tr -d ']['|cut -d/ -f2-)
  if [[ -n $LDSO ]]
  then
    echo found ld.so: $LDSO
  else
    echo ERROR: no dynamic loader of bootstrap rpm could be found, exiting
    exit 7
  fi
  echo trying to run bootstrap rpm
  export LD_LIBRARY_PATH="$BOOTSTRAP_CHROOTDIR"/lib64:"$BOOTSTRAP_CHROOTDIR"/lib:"$BOOTSTRAP_CHROOTDIR"/usr/lib64:"$BOOTSTRAP_CHROOTDIR"/usr/lib
  export RPM="$LDSO ""$BOOTSTRAP_CHROOTDIR"/bin/rpm
  $RPM -qa --root "$BOOTSTRAP_CHROOTDIR"
  RETVALUE=$?
  [[ $RETVALUE -ne 0 ]] && { echo ERROR: could not execute bootstrap rpm "$BOOTSTRAP_CHROOTDIR"/bin/rpm with detected loader, exiting; exit 8; }
fi

# finally the actual build
if [[ -z $NORPM ]]
then
  echo installing with \"rpm -i --root\"
  $RPM --initdb --root "$CHROOTDIR"
  $RPM -i --root "$CHROOTDIR" --ignorearch $RPMOPTS $PACKAGES_TO_INSTALL
else
  echo norpm was set, using rpm2cpio for unpacking
  for PKG in $PACKAGES_TO_INSTALL
  do
    echo $PKG
    if rpm2cpio $PKG |file -|grep -q xz.compressed
    then
      rpm2cpio $PKG | xz -d
    else
      rpm2cpio $PKG
    fi \
      | ( cd "$CHROOTDIR" && cpio -i --make-directories --preserve-modification-time --no-absolute-filenames --unconditional 2>/dev/null )
  done
fi
/bin/rm -rf "$CHROOTDIR"/var/lib/rpm/__db.00* "$BOOTSTRAP_CHROOTDIR"

COLORGREEN='\033[1;32m'
COLORNORMAL='\033[0;39m'
echo -e chroot environment build: $COLORGREEN $CHROOTDIR $COLORNORMAL with $COLORGREEN $(du -sm "$CHROOTDIR"|awk '{print$1}') MB $COLORNORMAL
ls -ld "$CHROOTDIR"/*

if [[ -n $MOUNT ]]
then
  echo mounting
  mkdir -p "$CHROOTDIR"/etc/yum.repos.d
  cd ..
  REPODATA=$(ls -d ./repodata 2>/dev/null;ls -d */repodata 2>/dev/null)
  REPODIRS=$(echo $REPODATA|sed 's,/repodata,,g')
  REPOBASEDIR=$(pwd)
  REPONAME=$(echo $REPOBASEDIR|sed s,.*/,,)
  for f in $REPODIRS
  do
    echo '
['$REPONAME'-'$f']
name='$REPONAME'-'$f'
baseurl=file://'$REPOBASEDIR'/'$f'
enabled=1
gpgcheck=0'
  done >"$CHROOTDIR"/etc/yum.repos.d/"$REPONAME".repo
  echo repo file written to "$CHROOTDIR"/etc/yum.repos.d/"$REPONAME".repo
  UMOUNT_COMMAND=""
  for f in "$REPOBASEDIR" /dev /proc
  do
    mkdir -p "$CHROOTDIR"/"$f"
    mount -o bind "$f" "$CHROOTDIR"/"$f"
    UMOUNT_COMMAND='umount '$CHROOTDIR'/'$f";$UMOUNT_COMMAND"
  done
  echo bindmounts done, umount command: $UMOUNT_COMMAND
fi

exit 0

